
#ifndef DISTRIBUTORBROADCASTCORE_HPP
#define DISTRIBUTORBROADCASTCORE_HPP 1


#include <string>
#include <sstream>
#include <vector>

#if !defined( _WIN32 ) && !defined( __clang__) && __GNUC__ <= 4 && __GNUC_MINOR__ <= 6
#include <tr1/memory>
#define SHARED_PTR std::tr1::shared_ptr
#else
#include <memory>
#define SHARED_PTR std::shared_ptr
#endif


#if defined(__linux__) || defined(__LINUX__)
#define HAVE_OPENSSL		1
#else
#define HAVE_OPENSSL		0
#endif

#if defined( __APPLE__) && defined( __MACH__)
#define HAVE_COMMONCRYPTO	1
#else
#define HAVE_COMMONCRYPTO	0
#endif


#if defined(__linux__) || defined(__LINUX__) || (defined( __APPLE__) && defined( __MACH__))
#define HAVE_POSIX			1
#include <sys/uio.h>
#else
#define HAVE_POSIX			0
#endif


#if defined(_WIN32) || defined(_WIN64)
#define HAVE_WINDOWSSECURITY	1
#define HAVE_WINSOCK			1
#include "json\json.h"
// ################ IOVEC-WSABUF ################
struct iovec {
	void*  iov_base;	// Starting address
	size_t iov_len;		// Number of bytes to transfer
};
typedef int ssize_t;
#else
#define HAVE_WINDOWSSECURITY	0
#define HAVE_WINSOCK			0
#include <json/json.h>
#endif


/**
\brief
Diese rein virtuelle Klasse gibt vor welche Funktionalitäten ein Benutzer der Klasse
DistributorBroadcastCore zur Verfügung stellen muss. Es sind die Call-Backs, die der
Distributor beim Eintreffen entsprechender Ereignisse aufruft. Typisch ruft der
Distributor die Funktionen aus einem separaten Thread-Kontext auf.

Der Path ist der Nachrichtenname.
Messages ohne Parameter (Value) haben einen Leerstring bzw. ein JSON-NULL als Parameter.
Die "Publication" sind Statusinformationen, die dem Endbenutzer angezeigt werden können.
Das Loging ist intern für Entwicklung und Support.
Ein Log-Level == 0 schaltet das Loging ab. 1 == ist normales Log, 2 und 3 für ein sehr ausführliches Log. 
Im Normalbetrieb 0 oder 1 verwenden.
*/
class DistributorCore_Adapter {
public:

	//! Dummy wegen Interface-Klasse.
	virtual ~DistributorCore_Adapter( void) {};

	//! Wird aufgerufen beim Eintreffen einer neuen Nachricht.
	virtual bool receive( const char* path, Json::Value value_j) = 0;
	//! Wird aufgerufen beim Abschalten des Distributors. (Auch bei Reinit.)
	virtual bool delete_publication( void) = 0;
	//! Neue Statusinformationen publizieren.
	virtual bool put_publication( Json::Value value_j) = 0;
	//! Abfrage nach aktuellem Log-Level
	virtual bool log_reached( int level) = 0;
	//! Log-Nachricht ausgeben.
	virtual void log_entry( int level, const std::string& entry) = 0;

};




/**
\brief
*/
class DistributorCore_Socket {
public:

	//! Dummy wegen Interface-Klasse.
	virtual ~DistributorCore_Socket( void) {};

	//! Socket öffnen
	virtual bool open( const std::vector< std::string>& def_ifaces, const std::string& iface, uint16_t port, std::string& self) = 0;
	//! Socket schließen
	virtual void close( void) = 0;
	//! Paket senden
	virtual ssize_t writev( const struct iovec* const iov, int iov_cnt) = 0;
	//! Paket emfangen, non-blocking
	virtual ssize_t recvfrom( void* buf, size_t len, std::string& from_addr) = 0;

};




/**
\brief
*/
class DistributorCore_Crypto {
public:

	static const int crypt_key_size = 16;	//!< AES128 => 128 Bit == 16 Byte
	static const int hmac_size = 32;		//!< SHA256 => 256 Bit == 32 Byte

	//! Dummy wegen Interface-Klasse.
	virtual ~DistributorCore_Crypto( void) {};

	//! Nach RFC2898 klartext Passphrase zu AES128 Schlüssel verwurschdeln. Z.B. OpenSSL::PKCS5_PBKDF2_HMAC_SHA1()
	virtual bool set_passphrase(
		const uint8_t* const passphrase,
		size_t passphrase_len,
		const uint8_t* const salt,
		size_t salt_len,
		int iter) = 0;
	//! hmac = SHA256( data + RFC2898( Passphrase)); Z.B. OpenSSL::HMAC() / OpenSSL::HMAC_Update()
	virtual bool make_hmac(
		const struct iovec* const data,
		unsigned int data_cnt,
		uint8_t* hmac) = 0;
	//! hmac == SHA256( data + RFC2898( Passphrase)); Z.B. OpenSSL::HMAC() / OpenSSL::HMAC_Update()
	virtual bool check_hmac(
		const uint8_t* const data,
		size_t len,
		const uint8_t* const hmac) = 0;
	//! AES128(). Z.B. OpenSSL::EVP_CipherUpdate()
	virtual size_t encrypt(
		const struct iovec* const clear_text,
		unsigned int clear_text_cnt,
		const uint8_t* const init_vect,
		uint8_t* cipher_text,
		size_t& cipher_text_len) = 0;
	//! AES128(). Z.B. OpenSSL::EVP_CipherUpdate()
	virtual bool decrypt(
		const uint8_t* const cipher_text,
		const uint8_t* const init_vect,
		uint8_t* const clear_text,
		size_t& len) = 0;

};




/**
\brief
Der Kernalgorithmus des Distributors.
*/
class DistributorBroadcastCore {
private:

	class Impl;
	SHARED_PTR< Impl> pimpl;

public:

	//! Instanzieren, aber noch nicht starten. Das macht init();
	DistributorBroadcastCore(  const size_t log_list_max, const Json::Value& system_config_j,
		DistributorCore_Adapter& adapter, DistributorCore_Crypto& crypto, DistributorCore_Socket& socket);
	//! Schluss, Aus, Ende.
	~DistributorBroadcastCore( void);
	//! Message senden / in sende-Warteschlange einreihen.
	void send( const std::string& path, const std::string& value);
	//! (Re-)Initialisieren / Benutzerkonfiguration laden, Distributor starten. Mindestens ein mal nach Konstruktor aufzurufen.
	bool init( const Json::Value& config_j);
	//! Statusinformationen neu zusammenstellen, aktualisieren und DistributorCore_Adapter::put_publication() aufrufen.
	void publish_status( void);

};




class LogEntry { //{{{1
private:

	DistributorCore_Adapter& adapter;
	int level;
	std::ostringstream stream;

public:

	LogEntry( DistributorCore_Adapter& adapt, int lev) :
		adapter( adapt),
		level( lev) {
	};


	~LogEntry( void) {
		adapter.log_entry( level, stream.str());
	};


	std::ostream&
	get_stream( void) {
		return stream;
	};

}; // }}}1




#endif // DISTRIBUTORBROADCASTCORE_HPP
