//           ///          //                                  C++ Cross Platform
//          /////        ////
//         /// XXX     XXX ///            ///////////   /////////     ///   ///
//        ///    XXX XXX    ///         ///             ///    ///   ///  ///
//       ///       XXX       ///         /////////     ///      //  //////
//      ///      XXX XXX      ///               ///   ///    ///   ///  ///
//     ////    XXX     XXX    ////    ////////////  //////////    ///   ///
//    ////                     ////
//   ////  M  O  B  O  T  I  X  ////////////////////////////////////////////////
//  //// Security Vision Systems //////////////////////////////////////////////
//
//
///////
     //
     //  MOBOTIX MxMessageSystem-SDK Demo
     //
     //  This software is licensed under the BSD licence,
     //  http://www.opensource.org/licenses/bsd-license.php
     //
     //  Copyright (c) 2005 - 2007, MOBOTIX AG
     //  All rights reserved.
     //
     //  Redistribution and use in source and binary forms, with or without
     //  modification, are permitted provided that the following conditions are
     //  met:
     //
     //  - Redistributions of source code must retain the above copyright
     //    notice, this list of conditions and the following disclaimer.
     //
     //  - Redistributions in binary form must reproduce the above copyright
     //    notice, this list of conditions and the following disclaimer in the
     //    documentation and/or other materials provided with the distribution.
     //
     //  - Neither the name of MOBOTIX AG nor the names of its contributors may
     //    be used to endorse or promote products derived from this software
     //    without specific prior written permission.
     //
     //  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
     //  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
     //  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
     //  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
     //  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
     //  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
     //  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
     //  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
     //  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
     //  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
     //  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
     //
///////

#include <fstream>
#include <iostream>


#include "DistributorBroadcastCore.hpp"

#if HAVE_OPENSSL > 0
#include "DistributorCore_CryptoOpenSSL.hpp"
#elif HAVE_COMMONCRYPTO > 0
#include "DistributorCore_CryptoCommonCrypto.hpp"
#elif HAVE_WINDOWSSECURITY > 0
#include "DistributorCore_CryptoWindowsSecurity.hpp"
#else
#error Neither OpenSSL nor CommonCrypto nor WindowsSecurity available!
#endif

#if HAVE_POSIX > 0
#include "DistributorCore_SocketPOSIX.hpp"
#elif HAVE_WINSOCK > 0
#include "DistributorCore_SocketWINSOCK.hpp"
#else
#error Neither POSIX nor WINSOCK available!
#endif


#define DEF_LOG_LIST_MAX		20


using namespace std;


static int debug;
static bool verbose;


class Adapter : // {{{1
public DistributorCore_Adapter {
public:

	Adapter( void);

	Adapter( const Adapter&) = delete;					//!< deleted copy constructor
	Adapter& operator=( const Adapter&) = delete;		//!< deleted copy assignment
	Adapter( Adapter&&) = delete;						//!< deleted move constructor
	Adapter& operator=( Adapter&&) = delete;			//!< deleted move assignment

	virtual bool receive( const char* path, Json::Value value_j);
	virtual bool delete_publication( void);
	virtual bool put_publication( Json::Value value_j);
	virtual bool log_reached( int level);
	virtual void log_entry( int level, const std::string& entry);

}; // class DistributorCore_Adapter }}}1


// class Adapter {{{1

Adapter::Adapter( void) {}


bool
Adapter::receive( const char* path, Json::Value value_j) { // {{{2
	if (verbose) {
		cout << endl << "receive: path=" << path << " value=" << value_j << endl;
	} else {
		cout << endl << path << "=" << value_j << endl;
	}
	return true;
} // Adapter::receive() }}}2


bool
Adapter::delete_publication( void) { // {{{2
	if (verbose) {
		cout << "delete_publication" << endl;
	}
	return true;
} // Adapter::delete_publication() }}}2


bool
Adapter::put_publication( Json::Value status_j) { // {{{2
	if (verbose) {
		cout << "put_publication=" << status_j << endl;
	}
	return true;
} // Adapter::put_publication() }}}2


bool
Adapter::log_reached( int level) { // {{{2
	return level <= debug;
} // Adapter::log_reached() }}}2


void
Adapter::log_entry( int level, const std::string& entry) { // {{{2
	cout << level << " " << entry << endl;
} // Adapter::log_entry() }}}2


// class Adapter }}}1


static bool
get_system_config( const char* system_config_fn, size_t& log_list_max, Json::Value& system_config_j) { // {{{1
	if (system_config_fn == NULL) {
		system_config_j[ "log_list_max"] = DEF_LOG_LIST_MAX;
		system_config_j[ "broadcast"][ "interfaces"][ 0] = "eth0";
	} else {
		ifstream system_config_st( system_config_fn);
		if (!system_config_st.is_open()) {
			cerr << "Can't open system configuration file!" << endl;
			return false;
		}
		Json::Reader reader;
		if (!reader.parse( system_config_st, system_config_j)) {
			cerr << "Can't parse system configuration file:" << reader.getFormattedErrorMessages() << endl;
			return false;
		}
	}
	log_list_max = DEF_LOG_LIST_MAX;
	const Json::Value& log_list_max_j = system_config_j[ "log_list_max"];
	if (log_list_max_j.isIntegral()) {
		log_list_max = log_list_max_j.asUInt();
	}
	if (log_list_max > 0 && log_list_max < 5) {
		log_list_max = 5;
	}
	if (log_list_max > 500) {
		log_list_max = 500;
	}
	return true;
} // }}}1


static bool
get_user_config( const char* user_config_fn, Json::Value& user_config_j) { // {{{1
	if (user_config_fn == NULL) {
//		user_config_j[ "broadcast"][ "iface"] = "eth1";
//		user_config_j[ "broadcast"][ "max_time_offset"] = 666;
		user_config_j[ "broadcast"][ "passphrase"] = "meinsm";
		user_config_j[ "broadcast"][ "poll_freq"] = 4;
		user_config_j[ "broadcast"][ "port"] = 19800;
		user_config_j[ "broadcast"][ "resend"] = 3;
	} else {
		ifstream user_config_st( user_config_fn);
		if (!user_config_st.is_open()) {
			cerr << "Can't open user configuration file!" << endl;
			return false;
		}
		Json::Reader reader;
		if (!reader.parse( user_config_st, user_config_j)) {
			cerr << "Can't parse user configuration file:" << reader.getFormattedErrorMessages() << endl;
			return false;
		}
	}
	return true;
} // }}}1


// MS Windows has no POSIX getopt(3)
static bool
getargs( // {{{1
	int argc,
	char** argv,
	const char*& system_config_fn,
	const char*& user_config_fn) {
	bool retval = true;
	for (int opt = 1; opt < argc; ++opt) {
		if (argv[ opt][ 0] != '-' ||
			argv[ opt][ 1] == '\0' ||
			argv[ opt][ 2] != '\0') {
			retval = false;
			break;
		}
		if (argv[ opt][ 1] == 'd' && opt + 1 < argc) {
			debug = atoi( argv[ ++opt]);
			continue;
		}
		if (argv[ opt][ 1] == 's' && opt + 1 < argc) {
			system_config_fn = argv[ ++opt];
			continue;
		}
		if (argv[ opt][ 1] == 'u' && opt + 1 < argc) {
			user_config_fn = argv[ ++opt];
			continue;
		}
		if (argv[ opt][ 1] == 'v') {
			verbose = true;
			continue;
		}
		retval = false;
		break;
	}
	if (!retval) {
		cout << argv[ 0] << " [-d debuglevel] [-s system_config.json] [-u user_config.json] [-v]" << endl;
	}
	return retval;
} // }}}1 getargs()


int
main( int argc, char** argv) { // {{{1
	debug = 0;
	verbose = false;
	const char* system_config_fn = NULL;
	const char* user_config_fn = NULL;
	if (!getargs( argc, argv, system_config_fn, user_config_fn)) {
		return EXIT_FAILURE;
	}
	size_t log_list_max;
	Json::Value system_config_j;
	if (!get_system_config( system_config_fn, log_list_max, system_config_j)) {
		return EXIT_FAILURE;
	}
	Json::Value user_config_j;
	if (!get_user_config( user_config_fn, user_config_j)) {
		return EXIT_FAILURE;
	}

	Adapter distri_adapter;
#if HAVE_OPENSSL > 0
	DistributorCore_CryptoOpenSSL distri_crypto( distri_adapter);
#elif HAVE_COMMONCRYPTO > 0
	DistributorCore_CryptoCommonCrypto distri_crypto( distri_adapter);
#elif HAVE_WINDOWSSECURITY > 0
	DistributorCore_CryptoWindowsSecurity distri_crypto( distri_adapter);
#endif
#if HAVE_POSIX > 0
	DistributorCore_SocketPOSIX distri_socket( distri_adapter);
#elif HAVE_WINSOCK > 0
	DistributorCore_SocketWINSOCK distri_socket( distri_adapter);
#endif
	DistributorBroadcastCore distri_core( log_list_max, system_config_j[ "broadcast"],
		distri_adapter, distri_crypto, distri_socket);
	if (!distri_core.init( user_config_j[ "broadcast"])) {
		cerr << "Initialization / Start failed!" << endl;
		return EXIT_FAILURE;
	}
	string path;
	string value;
	do {
		cout << "Enter path or '!' to exit or '?' to get statistics: ";
		getline( cin, path);
		if (path == "!") {
			break;
		}
		if (path == "?") {
			bool v = verbose;
			verbose = true;
			distri_core.publish_status();
			verbose = v;
			continue;
		}
		cout << "Enter value or '!' to exit or '?' to get statistics: ";
		getline( cin, value);
		if (value == "!") {
			break;
		}
		if (value.size() == 0) {
			value = "null";
		}
		if (path == "?") {
			bool v = verbose;
			verbose = true;
			distri_core.publish_status();
			verbose = v;
			continue;
		}
		cout << "sending path=" << path << endl << "sending value=" << value << endl;
		distri_core.send( path, value);
	} while (true);
	return EXIT_SUCCESS;
} // }}}1
